from __future__ import annotations
from typing import final, Literal
import datetime
import requests
from pydantic import TypeAdapter
from typing import Any
from ._base import APIError, ResponseValue, encode_path
from .types import (
    AcceptQuoteRequest,
    AmendOrderRequest,
    AmendOrderResponse,
    ApplySubaccountTransferRequest,
    ApplySubaccountTransferResponse,
    BatchCancelOrdersRequest,
    BatchCancelOrdersResponse,
    BatchCreateOrdersRequest,
    BatchCreateOrdersResponse,
    BatchGetMarketCandlesticksResponse,
    CancelOrderResponse,
    CreateApiKeyRequest,
    CreateApiKeyResponse,
    CreateMarketInMultivariateEventCollectionRequest,
    CreateMarketInMultivariateEventCollectionResponse,
    CreateOrderGroupRequest,
    CreateOrderGroupResponse,
    CreateOrderRequest,
    CreateOrderResponse,
    CreateQuoteRequest,
    CreateQuoteResponse,
    CreateRfqRequest,
    CreateRfqResponse,
    CreateSubaccountResponse,
    DecreaseOrderRequest,
    DecreaseOrderResponse,
    EmptyResponse,
    ExchangeStatus,
    GenerateApiKeyRequest,
    GenerateApiKeyResponse,
    GetApiKeysResponse,
    GetBalanceResponse,
    GetCommunicationsIdResponse,
    GetEventCandlesticksResponse,
    GetEventForecastPercentilesHistoryResponse,
    GetEventMetadataResponse,
    GetEventResponse,
    GetEventsResponse,
    GetExchangeAnnouncementsResponse,
    GetExchangeScheduleResponse,
    GetFillsResponse,
    GetFiltersBySportsResponse,
    GetIncentiveProgramsResponse,
    GetLiveDataResponse,
    GetLiveDatasResponse,
    GetMarketCandlesticksResponse,
    GetMarketOrderbookResponse,
    GetMarketResponse,
    GetMarketsResponse,
    GetMilestoneResponse,
    GetMilestonesResponse,
    GetMultivariateEventCollectionLookupHistoryResponse,
    GetMultivariateEventCollectionResponse,
    GetMultivariateEventCollectionsResponse,
    GetMultivariateEventsResponse,
    GetOrderGroupResponse,
    GetOrderGroupsResponse,
    GetOrderQueuePositionResponse,
    GetOrderQueuePositionsResponse,
    GetOrderResponse,
    GetOrdersResponse,
    GetPortfolioRestingOrderTotalValueResponse,
    GetPositionsResponse,
    GetQuoteResponse,
    GetQuotesResponse,
    GetRfQsResponse,
    GetRfqResponse,
    GetSeriesFeeChangesResponse,
    GetSeriesListResponse,
    GetSeriesResponse,
    GetSettlementsResponse,
    GetStructuredTargetResponse,
    GetStructuredTargetsResponse,
    GetSubaccountBalancesResponse,
    GetSubaccountTransfersResponse,
    GetTagsForSeriesCategoriesResponse,
    GetTradesResponse,
    GetUserDataTimestampResponse,
    LookupTickersForMarketInMultivariateEventCollectionRequest,
    LookupTickersForMarketInMultivariateEventCollectionResponse,
)


@final
class ApiKeysClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_api_keys(self) -> ResponseValue[GetApiKeysResponse]:
        """Get API Keys

        Endpoint for retrieving all API keys associated with the authenticated user.  API keys allow programmatic access to the platform without requiring username/password authentication. Each key has a unique identifier and name."""
        _path = "/api_keys"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetApiKeysResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_api_key(
        self, body: CreateApiKeyRequest
    ) -> ResponseValue[CreateApiKeyResponse]:
        """Create API Key

        Endpoint for creating a new API key with a user-provided public key.  This endpoint allows users with Premier or Market Maker API usage levels to create API keys by providing their own RSA public key. The platform will use this public key to verify signatures on API requests."""
        _path = "/api_keys"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(CreateApiKeyRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateApiKeyResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def generate_api_key(
        self, body: GenerateApiKeyRequest
    ) -> ResponseValue[GenerateApiKeyResponse]:
        """Generate API Key

        Endpoint for generating a new API key with an automatically created key pair.  This endpoint generates both a public and private RSA key pair. The public key is stored on the platform, while the private key is returned to the user and must be stored securely. The private key cannot be retrieved again."""
        _path = "/api_keys/generate"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(GenerateApiKeyRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GenerateApiKeyResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def delete_api_key(self, api_key: str) -> ResponseValue[None]:
        """Delete API Key

        Endpoint for deleting an existing API key.  This endpoint permanently deletes an API key. Once deleted, the key can no longer be used for authentication. This action cannot be undone."""
        _path = f"/api_keys/{encode_path(api_key)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )


@final
class CommunicationsClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_communications_id(self) -> ResponseValue[GetCommunicationsIdResponse]:
        """Get Communications ID

        Endpoint for getting the communications ID of the logged-in user."""
        _path = "/communications/id"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetCommunicationsIdResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_rf_qs(
        self,
        cursor: str | None = None,
        event_ticker: str | None = None,
        market_ticker: str | None = None,
        limit: int | None = None,
        status: str | None = None,
        creator_user_id: str | None = None,
    ) -> ResponseValue[GetRfQsResponse]:
        """Get RFQs

        Endpoint for getting RFQs"""
        _path = "/communications/rfqs"
        _params: dict[str, Any] = {
            "cursor": cursor,
            "event_ticker": event_ticker,
            "market_ticker": market_ticker,
            "limit": limit,
            "status": status,
            "creator_user_id": creator_user_id,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetRfQsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_rfq(self, body: CreateRfqRequest) -> ResponseValue[CreateRfqResponse]:
        """Create RFQ

        Endpoint for creating a new RFQ. You can have a maximum of 100 open RFQs at a time."""
        _path = "/communications/rfqs"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(CreateRfqRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateRfqResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_rfq(self, rfq_id: str) -> ResponseValue[GetRfqResponse]:
        """Get RFQ

        Endpoint for getting a single RFQ by id"""
        _path = f"/communications/rfqs/{encode_path(rfq_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetRfqResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def delete_rfq(self, rfq_id: str) -> ResponseValue[None]:
        """Delete RFQ

        Endpoint for deleting an RFQ by ID"""
        _path = f"/communications/rfqs/{encode_path(rfq_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )

    def get_quotes(
        self,
        cursor: str | None = None,
        event_ticker: str | None = None,
        market_ticker: str | None = None,
        limit: int | None = None,
        status: str | None = None,
        quote_creator_user_id: str | None = None,
        rfq_creator_user_id: str | None = None,
        rfq_creator_subtrader_id: str | None = None,
        rfq_id: str | None = None,
    ) -> ResponseValue[GetQuotesResponse]:
        """Get Quotes

        Endpoint for getting quotes"""
        _path = "/communications/quotes"
        _params: dict[str, Any] = {
            "cursor": cursor,
            "event_ticker": event_ticker,
            "market_ticker": market_ticker,
            "limit": limit,
            "status": status,
            "quote_creator_user_id": quote_creator_user_id,
            "rfq_creator_user_id": rfq_creator_user_id,
            "rfq_creator_subtrader_id": rfq_creator_subtrader_id,
            "rfq_id": rfq_id,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetQuotesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_quote(
        self, body: CreateQuoteRequest
    ) -> ResponseValue[CreateQuoteResponse]:
        """Create Quote

        Endpoint for creating a quote in response to an RFQ"""
        _path = "/communications/quotes"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(CreateQuoteRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateQuoteResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_quote(self, quote_id: str) -> ResponseValue[GetQuoteResponse]:
        """Get Quote

        Endpoint for getting a particular quote"""
        _path = f"/communications/quotes/{encode_path(quote_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetQuoteResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def delete_quote(self, quote_id: str) -> ResponseValue[None]:
        """Delete Quote

        Endpoint for deleting a quote, which means it can no longer be accepted."""
        _path = f"/communications/quotes/{encode_path(quote_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )

    def accept_quote(
        self, quote_id: str, body: AcceptQuoteRequest
    ) -> ResponseValue[None]:
        """Accept Quote

        Endpoint for accepting a quote. This will require the quoter to confirm"""
        _path = f"/communications/quotes/{encode_path(quote_id)}/accept"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="PUT",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(AcceptQuoteRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )

    def confirm_quote(
        self, quote_id: str, body: EmptyResponse | None = None
    ) -> ResponseValue[None]:
        """Confirm Quote

        Endpoint for confirming a quote. This will start a timer for order execution"""
        _path = f"/communications/quotes/{encode_path(quote_id)}/confirm"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="PUT",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(EmptyResponse).dump_json(body)
            if body is not None
            else None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )


@final
class EventsClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_market_candlesticks_by_event(
        self,
        ticker: str,
        series_ticker: str,
        start_ts: int,
        end_ts: int,
        period_interval: Literal[1, 60, 1440],
    ) -> ResponseValue[GetEventCandlesticksResponse]:
        """Get Event Candlesticks

        End-point for returning aggregated data across all markets corresponding to an event."""
        _path = f"/series/{encode_path(series_ticker)}/events/{encode_path(ticker)}/candlesticks"
        _params: dict[str, Any] = {
            "start_ts": start_ts,
            "end_ts": end_ts,
            "period_interval": period_interval,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetEventCandlesticksResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_events(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        with_nested_markets: bool | None = None,
        with_milestones: bool | None = None,
        status: Literal["open", "closed", "settled"] | None = None,
        series_ticker: str | None = None,
        min_close_ts: int | None = None,
    ) -> ResponseValue[GetEventsResponse]:
        """Get Events

        Get all events. This endpoint excludes multivariate events.
        To retrieve multivariate events, use the GET /events/multivariate endpoint.
        """
        _path = "/events"
        _params: dict[str, Any] = {
            "limit": limit,
            "cursor": cursor,
            "with_nested_markets": with_nested_markets,
            "with_milestones": with_milestones,
            "status": status,
            "series_ticker": series_ticker,
            "min_close_ts": min_close_ts,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetEventsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_multivariate_events(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        series_ticker: str | None = None,
        collection_ticker: str | None = None,
        with_nested_markets: bool | None = None,
    ) -> ResponseValue[GetMultivariateEventsResponse]:
        """Get Multivariate Events

        Retrieve multivariate (combo) events. These are dynamically created events from multivariate event collections. Supports filtering by series and collection ticker."""
        _path = "/events/multivariate"
        _params: dict[str, Any] = {
            "limit": limit,
            "cursor": cursor,
            "series_ticker": series_ticker,
            "collection_ticker": collection_ticker,
            "with_nested_markets": with_nested_markets,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMultivariateEventsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_event(
        self, event_ticker: str, with_nested_markets: bool | None = None
    ) -> ResponseValue[GetEventResponse]:
        """Get Event

        Endpoint for getting data about an event by its ticker.  An event represents a real-world occurrence that can be traded on, such as an election, sports game, or economic indicator release. Events contain one or more markets where users can place trades on different outcomes."""
        _path = f"/events/{encode_path(event_ticker)}"
        _params: dict[str, Any] = {"with_nested_markets": with_nested_markets}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetEventResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_event_metadata(
        self, event_ticker: str
    ) -> ResponseValue[GetEventMetadataResponse]:
        """Get Event Metadata

        Endpoint for getting metadata about an event by its ticker.  Returns only the metadata information for an event."""
        _path = f"/events/{encode_path(event_ticker)}/metadata"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetEventMetadataResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_event_forecast_percentiles_history(
        self,
        ticker: str,
        series_ticker: str,
        percentiles: list[int],
        start_ts: int,
        end_ts: int,
        period_interval: Literal[0, 1, 60, 1440],
    ) -> ResponseValue[GetEventForecastPercentilesHistoryResponse]:
        """Get Event Forecast Percentile History

        Endpoint for getting the historical raw and formatted forecast numbers for an event at specific percentiles."""
        _path = f"/series/{encode_path(series_ticker)}/events/{encode_path(ticker)}/forecast_percentile_history"
        _params: dict[str, Any] = {
            "percentiles": percentiles,
            "start_ts": start_ts,
            "end_ts": end_ts,
            "period_interval": period_interval,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetEventForecastPercentilesHistoryResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class ExchangeClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_exchange_status(self) -> ResponseValue[ExchangeStatus]:
        """Get Exchange Status

        Endpoint for getting the exchange status."""
        _path = "/exchange/status"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = ExchangeStatus.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_exchange_announcements(
        self,
    ) -> ResponseValue[GetExchangeAnnouncementsResponse]:
        """Get Exchange Announcements

        Endpoint for getting all exchange-wide announcements."""
        _path = "/exchange/announcements"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetExchangeAnnouncementsResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_series_fee_changes(
        self, series_ticker: str | None = None, show_historical: bool | None = None
    ) -> ResponseValue[GetSeriesFeeChangesResponse]:
        """Get Series Fee Changes"""
        _path = "/series/fee_changes"
        _params: dict[str, Any] = {
            "series_ticker": series_ticker,
            "show_historical": show_historical,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSeriesFeeChangesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_exchange_schedule(self) -> ResponseValue[GetExchangeScheduleResponse]:
        """Get Exchange Schedule

        Endpoint for getting the exchange schedule."""
        _path = "/exchange/schedule"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetExchangeScheduleResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_user_data_timestamp(self) -> ResponseValue[GetUserDataTimestampResponse]:
        """Get User Data Timestamp

        There is typically a short delay before exchange events are reflected in the API endpoints. Whenever possible, combine API responses to PUT/POST/DELETE requests with websocket data to obtain the most accurate view of the exchange state. This endpoint provides an approximate indication of when the data from the following endpoints was last validated: GetBalance, GetOrder(s), GetFills, GetPositions"""
        _path = "/exchange/user_data_timestamp"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetUserDataTimestampResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class FcmClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_fcm_orders(
        self,
        subtrader_id: str,
        cursor: str | None = None,
        event_ticker: str | None = None,
        ticker: str | None = None,
        min_ts: int | None = None,
        max_ts: int | None = None,
        status: Literal["resting", "canceled", "executed"] | None = None,
        limit: int | None = None,
    ) -> ResponseValue[GetOrdersResponse]:
        """Get FCM Orders

        Endpoint for FCM members to get orders filtered by subtrader ID.
        This endpoint requires FCM member access level and allows filtering orders by subtrader ID.
        """
        _path = "/fcm/orders"
        _params: dict[str, Any] = {
            "subtrader_id": subtrader_id,
            "cursor": cursor,
            "event_ticker": event_ticker,
            "ticker": ticker,
            "min_ts": min_ts,
            "max_ts": max_ts,
            "status": status,
            "limit": limit,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrdersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_fcm_positions(
        self,
        subtrader_id: str,
        ticker: str | None = None,
        event_ticker: str | None = None,
        count_filter: str | None = None,
        settlement_status: Literal["all", "unsettled", "settled"] | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> ResponseValue[GetPositionsResponse]:
        """Get FCM Positions

        Endpoint for FCM members to get market positions filtered by subtrader ID.
        This endpoint requires FCM member access level and allows filtering positions by subtrader ID.
        """
        _path = "/fcm/positions"
        _params: dict[str, Any] = {
            "subtrader_id": subtrader_id,
            "ticker": ticker,
            "event_ticker": event_ticker,
            "count_filter": count_filter,
            "settlement_status": settlement_status,
            "limit": limit,
            "cursor": cursor,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetPositionsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class IncentiveProgramsClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_incentive_programs(
        self,
        status: Literal["all", "active", "upcoming", "closed", "paid_out"]
        | None = None,
        type: Literal["all", "liquidity", "volume"] | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> ResponseValue[GetIncentiveProgramsResponse]:
        """Get Incentives

        List incentives with optional filters. Incentives are rewards programs for trading activity on specific markets."""
        _path = "/incentive_programs"
        _params: dict[str, Any] = {
            "status": status,
            "type": type,
            "limit": limit,
            "cursor": cursor,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetIncentiveProgramsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class LiveDataClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_live_data(
        self, type: str, milestone_id: str
    ) -> ResponseValue[GetLiveDataResponse]:
        """Get Live Data

        Get live data for a specific milestone"""
        _path = f"/live_data/{encode_path(type)}/milestone/{encode_path(milestone_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetLiveDataResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_live_datas(
        self, milestone_ids: list[str]
    ) -> ResponseValue[GetLiveDatasResponse]:
        """Get Multiple Live Data

        Get live data for multiple milestones"""
        _path = "/live_data/batch"
        _params: dict[str, Any] = {"milestone_ids": milestone_ids}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetLiveDatasResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class MarketClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_market_candlesticks(
        self,
        series_ticker: str,
        ticker: str,
        start_ts: int,
        end_ts: int,
        period_interval: Literal[1, 60, 1440],
        include_latest_before_start: bool | None = None,
    ) -> ResponseValue[GetMarketCandlesticksResponse]:
        """Get Market Candlesticks

        Time period length of each candlestick in minutes. Valid values: 1 (1 minute), 60 (1 hour), 1440 (1 day)."""
        _path = f"/series/{encode_path(series_ticker)}/markets/{encode_path(ticker)}/candlesticks"
        _params: dict[str, Any] = {
            "start_ts": start_ts,
            "end_ts": end_ts,
            "period_interval": period_interval,
            "include_latest_before_start": include_latest_before_start,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMarketCandlesticksResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_trades(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        ticker: str | None = None,
        min_ts: int | None = None,
        max_ts: int | None = None,
    ) -> ResponseValue[GetTradesResponse]:
        """Get Trades

        Endpoint for getting all trades for all markets.  A trade represents a completed transaction between two users on a specific market. Each trade includes the market ticker, price, quantity, and timestamp information.  This endpoint returns a paginated response. Use the 'limit' parameter to control page size (1-1000, defaults to 100). The response includes a 'cursor' field - pass this value in the 'cursor' parameter of your next request to get the next page. An empty cursor indicates no more pages are available."""
        _path = "/markets/trades"
        _params: dict[str, Any] = {
            "limit": limit,
            "cursor": cursor,
            "ticker": ticker,
            "min_ts": min_ts,
            "max_ts": max_ts,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetTradesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_market_orderbook(
        self, ticker: str, depth: int | None = None
    ) -> ResponseValue[GetMarketOrderbookResponse]:
        """Get Market Orderbook

        Endpoint for getting the current order book for a specific market.  The order book shows all active bid orders for both yes and no sides of a binary market. It returns yes bids and no bids only (no asks are returned). This is because in binary markets, a bid for yes at price X is equivalent to an ask for no at price (100-X). For example, a yes bid at 7¢ is the same as a no ask at 93¢, with identical contract sizes.  Each side shows price levels with their corresponding quantities and order counts, organized from best to worst prices."""
        _path = f"/markets/{encode_path(ticker)}/orderbook"
        _params: dict[str, Any] = {"depth": depth}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMarketOrderbookResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_series(
        self, series_ticker: str, include_volume: bool | None = None
    ) -> ResponseValue[GetSeriesResponse]:
        """Get Series

        Endpoint for getting data about a specific series by its ticker.  A series represents a template for recurring events that follow the same format and rules (e.g., "Monthly Jobs Report", "Weekly Initial Jobless Claims", "Daily Weather in NYC"). Series define the structure, settlement sources, and metadata that will be applied to each recurring event instance within that series."""
        _path = f"/series/{encode_path(series_ticker)}"
        _params: dict[str, Any] = {"include_volume": include_volume}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSeriesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_series_list(
        self,
        category: str | None = None,
        tags: str | None = None,
        include_product_metadata: bool | None = None,
        include_volume: bool | None = None,
    ) -> ResponseValue[GetSeriesListResponse]:
        """Get Series List

        Endpoint for getting data about multiple series with specified filters.  A series represents a template for recurring events that follow the same format and rules (e.g., "Monthly Jobs Report", "Weekly Initial Jobless Claims", "Daily Weather in NYC"). This endpoint allows you to browse and discover available series templates by category."""
        _path = "/series"
        _params: dict[str, Any] = {
            "category": category,
            "tags": tags,
            "include_product_metadata": include_product_metadata,
            "include_volume": include_volume,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSeriesListResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_markets(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        event_ticker: str | None = None,
        series_ticker: str | None = None,
        min_created_ts: int | None = None,
        max_created_ts: int | None = None,
        max_close_ts: int | None = None,
        min_close_ts: int | None = None,
        min_settled_ts: int | None = None,
        max_settled_ts: int | None = None,
        status: Literal["unopened", "open", "paused", "closed", "settled"]
        | None = None,
        tickers: str | None = None,
        mve_filter: Literal["only", "exclude"] | None = None,
    ) -> ResponseValue[GetMarketsResponse]:
        """Get Markets

        Filter by market status. Possible values: `unopened`, `open`, `closed`, `settled`. Leave empty to return markets with any status.
         - Only one `status` filter may be supplied at a time.
         - Timestamp filters will be mutually exclusive from other timestamp filters and certain status filters.

         | Compatible Timestamp Filters | Additional Status Filters|
         |------------------------------|--------------------------|
         | min_created_ts, max_created_ts | `unopened`, `open`, *empty* |
         | min_close_ts, max_close_ts | `closed`, *empty* |
         | min_settled_ts, max_settled_ts | `settled`, *empty* |
        """
        _path = "/markets"
        _params: dict[str, Any] = {
            "limit": limit,
            "cursor": cursor,
            "event_ticker": event_ticker,
            "series_ticker": series_ticker,
            "min_created_ts": min_created_ts,
            "max_created_ts": max_created_ts,
            "max_close_ts": max_close_ts,
            "min_close_ts": min_close_ts,
            "min_settled_ts": min_settled_ts,
            "max_settled_ts": max_settled_ts,
            "status": status,
            "tickers": tickers,
            "mve_filter": mve_filter,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMarketsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_market(self, ticker: str) -> ResponseValue[GetMarketResponse]:
        """Get Market

        Endpoint for getting data about a specific market by its ticker. A market represents a specific binary outcome within an event that users can trade on (e.g., "Will candidate X win?"). Markets have yes/no positions, current prices, volume, and settlement rules."""
        _path = f"/markets/{encode_path(ticker)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMarketResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def batch_get_market_candlesticks(
        self,
        market_tickers: str,
        start_ts: int,
        end_ts: int,
        period_interval: int,
        include_latest_before_start: bool | None = None,
    ) -> ResponseValue[BatchGetMarketCandlesticksResponse]:
        """Batch Get Market Candlesticks

        Endpoint for retrieving candlestick data for multiple markets.

        - Accepts up to 100 market tickers per request
        - Returns up to 10,000 candlesticks total across all markets
        - Returns candlesticks grouped by market_id
        - Optionally includes a synthetic initial candlestick for price continuity (see `include_latest_before_start` parameter)
        """
        _path = "/markets/candlesticks"
        _params: dict[str, Any] = {
            "market_tickers": market_tickers,
            "start_ts": start_ts,
            "end_ts": end_ts,
            "period_interval": period_interval,
            "include_latest_before_start": include_latest_before_start,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = BatchGetMarketCandlesticksResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class MilestoneClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_milestone(self, milestone_id: str) -> ResponseValue[GetMilestoneResponse]:
        """Get Milestone

        Endpoint for getting data about a specific milestone by its ID."""
        _path = f"/milestones/{encode_path(milestone_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMilestoneResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_milestones(
        self,
        limit: int,
        minimum_start_date: datetime.datetime | None = None,
        category: str | None = None,
        competition: str | None = None,
        source_id: str | None = None,
        type: str | None = None,
        related_event_ticker: str | None = None,
        cursor: str | None = None,
    ) -> ResponseValue[GetMilestonesResponse]:
        """Get Milestones

        Minimum start date to filter milestones. Format: RFC3339 timestamp"""
        _path = "/milestones"
        _params: dict[str, Any] = {
            "limit": limit,
            "minimum_start_date": minimum_start_date,
            "category": category,
            "competition": competition,
            "source_id": source_id,
            "type": type,
            "related_event_ticker": related_event_ticker,
            "cursor": cursor,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMilestonesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class MultivariateClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_multivariate_event_collection(
        self, collection_ticker: str
    ) -> ResponseValue[GetMultivariateEventCollectionResponse]:
        """Get Multivariate Event Collection

        Endpoint for getting data about a multivariate event collection by its ticker."""
        _path = f"/multivariate_event_collections/{encode_path(collection_ticker)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMultivariateEventCollectionResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_market_in_multivariate_event_collection(
        self,
        collection_ticker: str,
        body: CreateMarketInMultivariateEventCollectionRequest,
    ) -> ResponseValue[CreateMarketInMultivariateEventCollectionResponse]:
        """Create Market In Multivariate Event Collection

        Endpoint for creating an individual market in a multivariate event collection. This endpoint must be hit at least once before trading or looking up a market."""
        _path = f"/multivariate_event_collections/{encode_path(collection_ticker)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(
                CreateMarketInMultivariateEventCollectionRequest
            ).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateMarketInMultivariateEventCollectionResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_multivariate_event_collections(
        self,
        status: Literal["unopened", "open", "closed"] | None = None,
        associated_event_ticker: str | None = None,
        series_ticker: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> ResponseValue[GetMultivariateEventCollectionsResponse]:
        """Get Multivariate Event Collections

        Endpoint for getting data about multivariate event collections."""
        _path = "/multivariate_event_collections"
        _params: dict[str, Any] = {
            "status": status,
            "associated_event_ticker": associated_event_ticker,
            "series_ticker": series_ticker,
            "limit": limit,
            "cursor": cursor,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetMultivariateEventCollectionsResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_multivariate_event_collection_lookup_history(
        self, collection_ticker: str, lookback_seconds: Literal[10, 60, 300, 3600]
    ) -> ResponseValue[GetMultivariateEventCollectionLookupHistoryResponse]:
        """Get Multivariate Event Collection Lookup History

        Endpoint for retrieving which markets in an event collection were recently looked up."""
        _path = (
            f"/multivariate_event_collections/{encode_path(collection_ticker)}/lookup"
        )
        _params: dict[str, Any] = {"lookback_seconds": lookback_seconds}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = (
            GetMultivariateEventCollectionLookupHistoryResponse.model_validate_json(
                _response.content
            )
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def lookup_tickers_for_market_in_multivariate_event_collection(
        self,
        collection_ticker: str,
        body: LookupTickersForMarketInMultivariateEventCollectionRequest,
    ) -> ResponseValue[LookupTickersForMarketInMultivariateEventCollectionResponse]:
        """Lookup Tickers For Market In Multivariate Event Collection

        Endpoint for looking up an individual market in a multivariate event collection. If CreateMarketInMultivariateEventCollection has never been hit with that variable combination before, this will return a 404."""
        _path = (
            f"/multivariate_event_collections/{encode_path(collection_ticker)}/lookup"
        )
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="PUT",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(
                LookupTickersForMarketInMultivariateEventCollectionRequest
            ).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = LookupTickersForMarketInMultivariateEventCollectionResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class OrderGroupsClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_order_groups(self) -> ResponseValue[GetOrderGroupsResponse]:
        """Get Order Groups

        Retrieves all order groups for the authenticated user."""
        _path = "/portfolio/order_groups"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrderGroupsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_order_group(
        self, body: CreateOrderGroupRequest
    ) -> ResponseValue[CreateOrderGroupResponse]:
        """Create Order Group

        Creates a new order group with a contracts limit. When the limit is hit, all orders in the group are cancelled and no new orders can be placed until reset."""
        _path = "/portfolio/order_groups/create"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(CreateOrderGroupRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateOrderGroupResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_order_group(
        self, order_group_id: str
    ) -> ResponseValue[GetOrderGroupResponse]:
        """Get Order Group

        Retrieves details for a single order group including all order IDs and auto-cancel status."""
        _path = f"/portfolio/order_groups/{encode_path(order_group_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrderGroupResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def delete_order_group(self, order_group_id: str) -> ResponseValue[EmptyResponse]:
        """Delete Order Group

        Deletes an order group and cancels all orders within it. This permanently removes the group."""
        _path = f"/portfolio/order_groups/{encode_path(order_group_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = EmptyResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def reset_order_group(
        self, order_group_id: str, body: EmptyResponse | None = None
    ) -> ResponseValue[EmptyResponse]:
        """Reset Order Group

        Resets the order group's matched contracts counter to zero, allowing new orders to be placed again after the limit was hit."""
        _path = f"/portfolio/order_groups/{encode_path(order_group_id)}/reset"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="PUT",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(EmptyResponse).dump_json(body)
            if body is not None
            else None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = EmptyResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class OrdersClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_orders(
        self,
        ticker: str | None = None,
        event_ticker: str | None = None,
        min_ts: int | None = None,
        max_ts: int | None = None,
        status: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
        subaccount: int | None = None,
    ) -> ResponseValue[GetOrdersResponse]:
        """Get Orders

        Restricts the response to orders that have a certain status: resting, canceled, or executed."""
        _path = "/portfolio/orders"
        _params: dict[str, Any] = {
            "ticker": ticker,
            "event_ticker": event_ticker,
            "min_ts": min_ts,
            "max_ts": max_ts,
            "status": status,
            "limit": limit,
            "cursor": cursor,
            "subaccount": subaccount,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrdersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_order(
        self, body: CreateOrderRequest
    ) -> ResponseValue[CreateOrderResponse]:
        """Create Order

        Endpoint for submitting orders in a market. Each user is limited to 200 000 open orders at a time."""
        _path = "/portfolio/orders"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(CreateOrderRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateOrderResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_order(self, order_id: str) -> ResponseValue[GetOrderResponse]:
        """Get Order

        Endpoint for getting a single order."""
        _path = f"/portfolio/orders/{encode_path(order_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrderResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def cancel_order(self, order_id: str) -> ResponseValue[CancelOrderResponse]:
        """Cancel Order

        Endpoint for canceling orders. The value for the orderId should match the id field of the order you want to decrease. Commonly, DELETE-type endpoints return 204 status with no body content on success. But we can't completely delete the order, as it may be partially filled already. Instead, the DeleteOrder endpoint reduce the order completely, essentially zeroing the remaining resting contracts on it. The zeroed order is returned on the response payload as a form of validation for the client."""
        _path = f"/portfolio/orders/{encode_path(order_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CancelOrderResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def batch_create_orders(
        self, body: BatchCreateOrdersRequest
    ) -> ResponseValue[BatchCreateOrdersResponse]:
        """Batch Create Orders

        Endpoint for submitting a batch of orders. Each order in the batch is counted against the total rate limit for order operations. Consequently, the size of the batch is capped by the current per-second rate-limit configuration applicable to the user. At the moment of writing, the limit is 20 orders per batch."""
        _path = "/portfolio/orders/batched"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(BatchCreateOrdersRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = BatchCreateOrdersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def batch_cancel_orders(
        self, body: BatchCancelOrdersRequest
    ) -> ResponseValue[BatchCancelOrdersResponse]:
        """Batch Cancel Orders

        Endpoint for cancelling up to 20 orders at once."""
        _path = "/portfolio/orders/batched"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="DELETE",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(BatchCancelOrdersRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = BatchCancelOrdersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def amend_order(
        self, order_id: str, body: AmendOrderRequest
    ) -> ResponseValue[AmendOrderResponse]:
        """Amend Order

        Endpoint for amending the max number of fillable contracts and/or price in an existing order. Max fillable contracts is `remaining_count` + `fill_count`."""
        _path = f"/portfolio/orders/{encode_path(order_id)}/amend"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(AmendOrderRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = AmendOrderResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def decrease_order(
        self, order_id: str, body: DecreaseOrderRequest
    ) -> ResponseValue[DecreaseOrderResponse]:
        """Decrease Order

        Endpoint for decreasing the number of contracts in an existing order. This is the only kind of edit available on order quantity. Cancelling an order is equivalent to decreasing an order amount to zero."""
        _path = f"/portfolio/orders/{encode_path(order_id)}/decrease"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(DecreaseOrderRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = DecreaseOrderResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_order_queue_positions(
        self, market_tickers: str | None = None, event_ticker: str | None = None
    ) -> ResponseValue[GetOrderQueuePositionsResponse]:
        """Get Queue Positions for Orders

        Endpoint for getting queue positions for all resting orders. Queue position represents the number of contracts that need to be matched before an order receives a partial or full match, determined using price-time priority."""
        _path = "/portfolio/orders/queue_positions"
        _params: dict[str, Any] = {
            "market_tickers": market_tickers,
            "event_ticker": event_ticker,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrderQueuePositionsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_order_queue_position(
        self, order_id: str
    ) -> ResponseValue[GetOrderQueuePositionResponse]:
        """Get Order Queue Position

        Endpoint for getting an order's queue position in the order book. This represents the amount of orders that need to be matched before this order receives a partial or full match. Queue position is determined using a price-time priority."""
        _path = f"/portfolio/orders/{encode_path(order_id)}/queue_position"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetOrderQueuePositionResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class PortfolioClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_balance(self) -> ResponseValue[GetBalanceResponse]:
        """Get Balance

        Endpoint for getting the balance and portfolio value of a member. Both values are returned in cents."""
        _path = "/portfolio/balance"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetBalanceResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def create_subaccount(self) -> ResponseValue[CreateSubaccountResponse]:
        """Create Subaccount

        Creates a new subaccount for the authenticated user. Subaccounts are numbered sequentially starting from 1. Maximum 32 subaccounts per user."""
        _path = "/portfolio/subaccounts"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = CreateSubaccountResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def apply_subaccount_transfer(
        self, body: ApplySubaccountTransferRequest
    ) -> ResponseValue[ApplySubaccountTransferResponse]:
        """Transfer Between Subaccounts

        Transfers funds between the authenticated user's subaccounts. Use 0 for the primary account, or 1-32 for numbered subaccounts."""
        _path = "/portfolio/subaccounts/transfer"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(ApplySubaccountTransferRequest).dump_json(body),
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = ApplySubaccountTransferResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_subaccount_balances(self) -> ResponseValue[GetSubaccountBalancesResponse]:
        """Get All Subaccount Balances

        Gets balances for all subaccounts including the primary account."""
        _path = "/portfolio/subaccounts/balances"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSubaccountBalancesResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_subaccount_transfers(
        self, limit: int | None = None, cursor: str | None = None
    ) -> ResponseValue[GetSubaccountTransfersResponse]:
        """Get Subaccount Transfers

        Gets a paginated list of all transfers between subaccounts for the authenticated user."""
        _path = "/portfolio/subaccounts/transfers"
        _params: dict[str, Any] = {"limit": limit, "cursor": cursor}
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSubaccountTransfersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_positions(
        self,
        cursor: str | None = None,
        limit: int | None = None,
        count_filter: str | None = None,
        ticker: str | None = None,
        event_ticker: str | None = None,
        subaccount: int | None = None,
    ) -> ResponseValue[GetPositionsResponse]:
        """Get Positions

        Restricts the positions to those with any of following fields with non-zero values, as a comma separated list. The following values are accepted: position, total_traded"""
        _path = "/portfolio/positions"
        _params: dict[str, Any] = {
            "cursor": cursor,
            "limit": limit,
            "count_filter": count_filter,
            "ticker": ticker,
            "event_ticker": event_ticker,
            "subaccount": subaccount,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetPositionsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_settlements(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        ticker: str | None = None,
        event_ticker: str | None = None,
        min_ts: int | None = None,
        max_ts: int | None = None,
    ) -> ResponseValue[GetSettlementsResponse]:
        """Get Settlements

        Endpoint for getting the member's settlements historical track."""
        _path = "/portfolio/settlements"
        _params: dict[str, Any] = {
            "limit": limit,
            "cursor": cursor,
            "ticker": ticker,
            "event_ticker": event_ticker,
            "min_ts": min_ts,
            "max_ts": max_ts,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetSettlementsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_portfolio_resting_order_total_value(
        self,
    ) -> ResponseValue[GetPortfolioRestingOrderTotalValueResponse]:
        """Get Total Resting Order Value

        Endpoint for getting the total value, in cents, of resting orders. This endpoint is only intended for use by FCM members (rare). Note: If you're uncertain about this endpoint, it likely does not apply to you."""
        _path = "/portfolio/summary/total_resting_order_value"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetPortfolioRestingOrderTotalValueResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_fills(
        self,
        ticker: str | None = None,
        order_id: str | None = None,
        min_ts: int | None = None,
        max_ts: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
        subaccount: int | None = None,
    ) -> ResponseValue[GetFillsResponse]:
        """Get Fills

        Endpoint for getting all fills for the member. A fill is when a trade you have is matched."""
        _path = "/portfolio/fills"
        _params: dict[str, Any] = {
            "ticker": ticker,
            "order_id": order_id,
            "min_ts": min_ts,
            "max_ts": max_ts,
            "limit": limit,
            "cursor": cursor,
            "subaccount": subaccount,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetFillsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class SearchClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_tags_for_series_categories(
        self,
    ) -> ResponseValue[GetTagsForSeriesCategoriesResponse]:
        """Get Tags for Series Categories

        Retrieve tags organized by series categories.

        This endpoint returns a mapping of series categories to their associated tags, which can be used for filtering and search functionality.
        """
        _path = "/search/tags_by_categories"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetTagsForSeriesCategoriesResponse.model_validate_json(
            _response.content
        )
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_filters_for_sports(self) -> ResponseValue[GetFiltersBySportsResponse]:
        """Get Filters for Sports

        Retrieve available filters organized by sport.

        This endpoint returns filtering options available for each sport, including scopes and competitions. It also provides an ordered list of sports for display purposes.
        """
        _path = "/search/filters_by_sport"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetFiltersBySportsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class StructuredTargetsClient:
    def __init__(self, session: requests.Session, base_url: str) -> None:
        self._session = session
        self._base_url = base_url

    def get_structured_targets(
        self,
        type: str | None = None,
        competition: str | None = None,
        page_size: int | None = None,
        cursor: str | None = None,
    ) -> ResponseValue[GetStructuredTargetsResponse]:
        """Get Structured Targets

        Page size (min: 1, max: 2000)"""
        _path = "/structured_targets"
        _params: dict[str, Any] = {
            "type": type,
            "competition": competition,
            "page_size": page_size,
            "cursor": cursor,
        }
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetStructuredTargetsResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_structured_target(
        self, structured_target_id: str
    ) -> ResponseValue[GetStructuredTargetResponse]:
        """Get Structured Target

        Endpoint for getting data about a specific structured target by its ID."""
        _path = f"/structured_targets/{encode_path(structured_target_id)}"
        _url = self._base_url.rstrip("/") + _path
        _response = self._session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = GetStructuredTargetResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )


@final
class Client:
    api_keys: "ApiKeysClient"
    communications: "CommunicationsClient"
    events: "EventsClient"
    exchange: "ExchangeClient"
    fcm: "FcmClient"
    incentive_programs: "IncentiveProgramsClient"
    live_data: "LiveDataClient"
    market: "MarketClient"
    milestone: "MilestoneClient"
    multivariate: "MultivariateClient"
    order_groups: "OrderGroupsClient"
    orders: "OrdersClient"
    portfolio: "PortfolioClient"
    search: "SearchClient"
    structured_targets: "StructuredTargetsClient"

    def __init__(self, base_url: str) -> None:
        self._base_url = base_url
        self._session = requests.Session()
        self.api_keys = ApiKeysClient(session=self._session, base_url=self._base_url)
        self.communications = CommunicationsClient(
            session=self._session, base_url=self._base_url
        )
        self.events = EventsClient(session=self._session, base_url=self._base_url)
        self.exchange = ExchangeClient(session=self._session, base_url=self._base_url)
        self.fcm = FcmClient(session=self._session, base_url=self._base_url)
        self.incentive_programs = IncentiveProgramsClient(
            session=self._session, base_url=self._base_url
        )
        self.live_data = LiveDataClient(session=self._session, base_url=self._base_url)
        self.market = MarketClient(session=self._session, base_url=self._base_url)
        self.milestone = MilestoneClient(session=self._session, base_url=self._base_url)
        self.multivariate = MultivariateClient(
            session=self._session, base_url=self._base_url
        )
        self.order_groups = OrderGroupsClient(
            session=self._session, base_url=self._base_url
        )
        self.orders = OrdersClient(session=self._session, base_url=self._base_url)
        self.portfolio = PortfolioClient(session=self._session, base_url=self._base_url)
        self.search = SearchClient(session=self._session, base_url=self._base_url)
        self.structured_targets = StructuredTargetsClient(
            session=self._session, base_url=self._base_url
        )

    def close(self) -> None:
        self._session.close()

    def __enter__(self) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        self.close()



"""aiohttp backend for HTTP client generation."""

from __future__ import annotations

import libcst as cst

from senzo.backends.base import HttpBackend
from senzo.codegen.utils import make_annotation, make_docstring, make_param
from senzo.operation import (
    ContentType,
    OperationMethod,
    OperationParameter,
    OperationResponse,
    ParameterLocation,
    WebSocketOperation,
)


class AiohttpBackend(HttpBackend):
    """aiohttp backend - async only.

    Generates fully functional async API client methods using the aiohttp library.
    This backend only supports asynchronous operations.

    Key differences from httpx:
    - Uses aiohttp.ClientSession instead of httpx.AsyncClient
    - Response body must be read with await resp.read() before decoding
    - Session is created in __aenter__ (lazy initialization)
    """

    def __init__(self) -> None:
        super().__init__(async_mode=True)

    def generate_client_class(
        self,
        name: str,
        base_url_param: bool = True,
    ) -> cst.ClassDef:
        """Generate the client class with constructor."""
        init_body: list[cst.BaseStatement] = []

        # self._base_url = base_url
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_base_url"),
                                )
                            )
                        ],
                        value=cst.Name("base_url"),
                    )
                ]
            )
        )

        # self._session: aiohttp.ClientSession | None = None (lazy init)
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.AnnAssign(
                        target=cst.Attribute(
                            value=cst.Name("self"),
                            attr=cst.Name("_session"),
                        ),
                        annotation=make_annotation("aiohttp.ClientSession | None"),
                        value=cst.Name("None"),
                    )
                ]
            )
        )

        # Build init params - add inner if hooks configured
        init_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("base_url"),
                annotation=cst.Annotation(annotation=cst.Name("str")),
            ),
        ]

        # Add inner parameter if hooks are configured
        if self._hooks_config is not None:
            inner_class = self._hooks_config.get_inner_class()
            init_params.append(
                cst.Param(
                    name=cst.Name("inner"),
                    annotation=cst.Annotation(annotation=cst.Name(inner_class)),
                )
            )
            # self._inner = inner
            init_body.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[
                                cst.AssignTarget(
                                    target=cst.Attribute(
                                        value=cst.Name("self"),
                                        attr=cst.Name("_inner"),
                                    )
                                )
                            ],
                            value=cst.Name("inner"),
                        )
                    ]
                )
            )

        init_func = cst.FunctionDef(
            name=cst.Name("__init__"),
            params=cst.Parameters(params=init_params),
            body=cst.IndentedBlock(body=init_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
        )

        # Generate close method
        close_body: list[cst.BaseStatement] = [
            cst.If(
                test=cst.Comparison(
                    left=cst.Attribute(
                        value=cst.Name("self"),
                        attr=cst.Name("_session"),
                    ),
                    comparisons=[
                        cst.ComparisonTarget(
                            operator=cst.IsNot(),
                            comparator=cst.Name("None"),
                        )
                    ],
                ),
                body=cst.IndentedBlock(
                    body=[
                        cst.SimpleStatementLine(
                            body=[
                                cst.Expr(
                                    value=cst.Await(
                                        expression=cst.Call(
                                            func=cst.Attribute(
                                                value=cst.Attribute(
                                                    value=cst.Name("self"),
                                                    attr=cst.Name("_session"),
                                                ),
                                                attr=cst.Name("close"),
                                            ),
                                            args=[],
                                        )
                                    )
                                )
                            ]
                        ),
                        cst.SimpleStatementLine(
                            body=[
                                cst.Assign(
                                    targets=[
                                        cst.AssignTarget(
                                            target=cst.Attribute(
                                                value=cst.Name("self"),
                                                attr=cst.Name("_session"),
                                            )
                                        )
                                    ],
                                    value=cst.Name("None"),
                                )
                            ]
                        ),
                    ]
                ),
                orelse=None,
            )
        ]

        close_func = cst.FunctionDef(
            name=cst.Name("close"),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=close_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
            asynchronous=cst.Asynchronous(),
        )

        # Async context manager methods
        # __aenter__ creates the session lazily
        enter_body: list[cst.BaseStatement] = [
            cst.If(
                test=cst.Comparison(
                    left=cst.Attribute(
                        value=cst.Name("self"),
                        attr=cst.Name("_session"),
                    ),
                    comparisons=[
                        cst.ComparisonTarget(
                            operator=cst.Is(),
                            comparator=cst.Name("None"),
                        )
                    ],
                ),
                body=cst.IndentedBlock(
                    body=[
                        cst.SimpleStatementLine(
                            body=[
                                cst.Assign(
                                    targets=[
                                        cst.AssignTarget(
                                            target=cst.Attribute(
                                                value=cst.Name("self"),
                                                attr=cst.Name("_session"),
                                            )
                                        )
                                    ],
                                    value=cst.Call(
                                        func=cst.Attribute(
                                            value=cst.Name("aiohttp"),
                                            attr=cst.Name("ClientSession"),
                                        ),
                                        args=[
                                            cst.Arg(
                                                keyword=cst.Name("base_url"),
                                                value=cst.Attribute(
                                                    value=cst.Name("self"),
                                                    attr=cst.Name("_base_url"),
                                                ),
                                            )
                                        ],
                                    ),
                                )
                            ]
                        )
                    ]
                ),
                orelse=None,
            ),
            cst.SimpleStatementLine(body=[cst.Return(value=cst.Name("self"))]),
        ]

        exit_body = [
            cst.SimpleStatementLine(
                body=[
                    cst.Expr(
                        value=cst.Await(
                            expression=cst.Call(
                                func=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("close"),
                                ),
                                args=[],
                            )
                        )
                    )
                ]
            )
        ]

        self_type = f'"{name}"'

        enter_func = cst.FunctionDef(
            name=cst.Name("__aenter__"),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=enter_body),
            returns=make_annotation(self_type),
            asynchronous=cst.Asynchronous(),
        )

        exit_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("exc_type"),
                annotation=make_annotation("type[BaseException] | None"),
            ),
            cst.Param(
                name=cst.Name("exc_val"),
                annotation=make_annotation("BaseException | None"),
            ),
            cst.Param(
                name=cst.Name("exc_tb"),
                annotation=make_annotation("object"),
            ),
        ]

        exit_func = cst.FunctionDef(
            name=cst.Name("__aexit__"),
            params=cst.Parameters(params=exit_params),
            body=cst.IndentedBlock(body=exit_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
            asynchronous=cst.Asynchronous(),
        )

        # _ensure_session helper for methods called outside context manager
        ensure_body: list[cst.BaseStatement] = [
            cst.If(
                test=cst.Comparison(
                    left=cst.Attribute(
                        value=cst.Name("self"),
                        attr=cst.Name("_session"),
                    ),
                    comparisons=[
                        cst.ComparisonTarget(
                            operator=cst.Is(),
                            comparator=cst.Name("None"),
                        )
                    ],
                ),
                body=cst.IndentedBlock(
                    body=[
                        cst.SimpleStatementLine(
                            body=[
                                cst.Raise(
                                    exc=cst.Call(
                                        func=cst.Name("RuntimeError"),
                                        args=[
                                            cst.Arg(
                                                value=cst.SimpleString(
                                                    '"Client must be used as async context manager"'
                                                )
                                            )
                                        ],
                                    )
                                )
                            ]
                        )
                    ]
                ),
                orelse=None,
            ),
            cst.SimpleStatementLine(
                body=[
                    cst.Return(
                        value=cst.Attribute(
                            value=cst.Name("self"),
                            attr=cst.Name("_session"),
                        )
                    )
                ]
            ),
        ]

        ensure_func = cst.FunctionDef(
            name=cst.Name("_ensure_session"),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=ensure_body),
            returns=make_annotation("aiohttp.ClientSession"),
        )

        class_body: list[cst.BaseStatement] = [
            init_func,
            close_func,
            enter_func,
            exit_func,
            ensure_func,
        ]

        return cst.ClassDef(
            name=cst.Name(name),
            body=cst.IndentedBlock(body=class_body),
            decorators=[cst.Decorator(decorator=cst.Name("final"))],
        )

    def generate_sub_client_class(
        self,
        name: str,
    ) -> cst.ClassDef:
        """Generate a sub-client class that shares an existing session.

        Sub-clients receive a reference to the parent client and access the session
        through it. This handles aiohttp's lazy session creation in __aenter__.

        We also store _base_url for building full URLs in hooks.
        """
        init_body: list[cst.BaseStatement] = []

        # self._parent = parent
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_parent"),
                                )
                            )
                        ],
                        value=cst.Name("parent"),
                    )
                ]
            )
        )

        # self._base_url = base_url
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_base_url"),
                                )
                            )
                        ],
                        value=cst.Name("base_url"),
                    )
                ]
            )
        )

        # Build init params - use Any for parent type to avoid circular reference
        init_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("parent"),
                annotation=make_annotation("Any"),
            ),
            cst.Param(
                name=cst.Name("base_url"),
                annotation=cst.Annotation(annotation=cst.Name("str")),
            ),
        ]

        # Add inner parameter if hooks are configured
        if self._hooks_config is not None:
            inner_class = self._hooks_config.get_inner_class()
            init_params.append(
                cst.Param(
                    name=cst.Name("inner"),
                    annotation=cst.Annotation(annotation=cst.Name(inner_class)),
                )
            )
            # self._inner = inner
            init_body.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[
                                cst.AssignTarget(
                                    target=cst.Attribute(
                                        value=cst.Name("self"),
                                        attr=cst.Name("_inner"),
                                    )
                                )
                            ],
                            value=cst.Name("inner"),
                        )
                    ]
                )
            )

        init_func = cst.FunctionDef(
            name=cst.Name("__init__"),
            params=cst.Parameters(params=init_params),
            body=cst.IndentedBlock(body=init_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
        )

        # Sub-clients access the session through the parent client.
        # This delegates to parent's _ensure_session which handles the None check.
        ensure_body: list[cst.BaseStatement] = [
            cst.SimpleStatementLine(
                body=[
                    cst.Return(
                        value=cst.Call(
                            func=cst.Attribute(
                                value=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_parent"),
                                ),
                                attr=cst.Name("_ensure_session"),
                            ),
                            args=[],
                        )
                    )
                ]
            ),
        ]

        ensure_func = cst.FunctionDef(
            name=cst.Name("_ensure_session"),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=ensure_body),
            returns=make_annotation("aiohttp.ClientSession"),
        )

        class_body: list[cst.BaseStatement] = [init_func, ensure_func]

        return cst.ClassDef(
            name=cst.Name(name),
            body=cst.IndentedBlock(body=class_body),
            decorators=[cst.Decorator(decorator=cst.Name("final"))],
        )

    def generate_method(
        self,
        operation: OperationMethod,
    ) -> cst.FunctionDef:
        """Generate an API method with full request/response handling."""
        method_name = operation.python_method_name()
        params = self._build_params(operation)
        body = self._build_method_body(operation)
        return_type = self._get_return_type(operation)

        decorators: list[cst.Decorator] = []
        if operation.deprecated:
            decorators.append(
                cst.Decorator(
                    decorator=cst.Call(
                        func=cst.Name("deprecated"),
                        args=[
                            cst.Arg(
                                value=cst.SimpleString(
                                    f'"{operation.operation_id} is deprecated"'
                                )
                            )
                        ],
                    )
                )
            )

        func = cst.FunctionDef(
            name=cst.Name(method_name),
            params=cst.Parameters(params=params),
            body=cst.IndentedBlock(body=body),
            returns=make_annotation(return_type),
            asynchronous=cst.Asynchronous(),
            decorators=decorators if decorators else [],
        )

        # Add docstring if summary or description exists
        if operation.summary or operation.description:
            doc_text = operation.summary or ""
            if operation.description and operation.description != operation.summary:
                if doc_text:
                    doc_text += "\n\n"
                doc_text += operation.description
            doc_stmt = make_docstring(doc_text)
            new_body = [doc_stmt, *body]
            func = func.with_changes(body=cst.IndentedBlock(body=new_body))

        return func

    def _build_params(
        self,
        operation: OperationMethod,
    ) -> list[cst.Param]:
        """Build function parameters from operation parameters."""
        params: list[cst.Param] = [cst.Param(name=cst.Name("self"))]

        # Add required parameters first
        for param in operation.parameters:
            if param.required:
                params.append(
                    make_param(param.name, param.type_annotation, default=None)
                )

        # Add request body if required
        if operation.request_body and operation.request_body.required:
            params.append(
                make_param("body", operation.request_body.type_annotation, default=None)
            )

        # Add optional parameters
        for param in operation.parameters:
            if not param.required:
                params.append(
                    make_param(
                        param.name,
                        param.type_annotation,
                        default=cst.Name("None"),
                    )
                )

        # Add optional request body
        if operation.request_body and not operation.request_body.required:
            params.append(
                make_param(
                    "body",
                    f"{operation.request_body.type_annotation} | None",
                    default=cst.Name("None"),
                )
            )

        return params

    def _build_method_body(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build the method body with request construction and response handling."""
        statements: list[cst.BaseStatement] = []

        # Build OperationInfo only if hooks are configured
        if self._hooks_config is not None:
            statements.append(self._build_operation_info(operation))

        # Build URL path
        path_expr = self._build_path_expression(operation)
        statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[cst.AssignTarget(target=cst.Name("_path"))],
                        value=path_expr,
                    )
                ]
            )
        )

        # Build query params
        query_params = [
            p for p in operation.parameters if p.location == ParameterLocation.QUERY
        ]
        if query_params:
            statements.append(self._build_query_params(query_params))

        # Build headers
        header_params = [
            p for p in operation.parameters if p.location == ParameterLocation.HEADER
        ]
        if header_params:
            statements.append(self._build_header_params(header_params))

        # Build request execution
        if self._hooks_config is not None:
            # With hooks: build RequestData, call hook functions
            statements.append(
                self._build_request_data(operation, query_params, header_params)
            )
            statements.extend(self._build_hooked_execution(operation))
        else:
            # Without hooks: direct HTTP call
            statements.extend(
                self._build_direct_execution(operation, query_params, header_params)
            )

        # Handle response
        statements.extend(self._build_response_handling(operation))

        return statements

    def _build_operation_info(
        self,
        operation: OperationMethod,
    ) -> cst.SimpleStatementLine:
        """Build OperationInfo dataclass instantiation."""
        tags_list = cst.List(
            elements=[
                cst.Element(cst.SimpleString(f'"{tag}"')) for tag in operation.tags
            ]
        )
        security_list = cst.List(
            elements=[
                cst.Element(cst.SimpleString(f'"{scheme}"'))
                for scheme in operation.security
            ]
        )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_info"))],
                    value=cst.Call(
                        func=cst.Name("OperationInfo"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("operation_id"),
                                value=cst.SimpleString(f'"{operation.operation_id}"'),
                            ),
                            cst.Arg(
                                keyword=cst.Name("method"),
                                value=cst.SimpleString(
                                    f'"{operation.method.value.upper()}"'
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("path_template"),
                                value=cst.SimpleString(f'"{operation.path.template}"'),
                            ),
                            cst.Arg(
                                keyword=cst.Name("tags"),
                                value=tags_list,
                            ),
                            cst.Arg(
                                keyword=cst.Name("security"),
                                value=security_list,
                            ),
                        ],
                    ),
                )
            ]
        )

    def _build_request_data(
        self,
        operation: OperationMethod,
        query_params: list[OperationParameter],
        header_params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build RequestData dataclass instantiation."""
        # URL: self._base_url.rstrip("/") + _path
        # This handles both base URLs with and without trailing slashes
        url_expr = cst.BinaryOperation(
            left=cst.Call(
                func=cst.Attribute(
                    value=cst.Attribute(
                        value=cst.Name("self"),
                        attr=cst.Name("_base_url"),
                    ),
                    attr=cst.Name("rstrip"),
                ),
                args=[cst.Arg(value=cst.SimpleString('"/"'))],
            ),
            operator=cst.Add(),
            right=cst.Name("_path"),
        )

        # Headers dict
        if header_params:
            headers_expr = cst.Name("_headers")
        else:
            headers_expr = cst.Dict(elements=[])

        # Add Content-Type header for JSON body
        if (
            operation.request_body
            and operation.request_body.content_type == ContentType.JSON
        ):
            if header_params:
                # Merge with existing headers
                headers_expr = cst.DictComp(
                    key=cst.Name("k"),
                    value=cst.Name("v"),
                    for_in=cst.CompFor(
                        target=cst.Tuple(
                            elements=[
                                cst.Element(cst.Name("k")),
                                cst.Element(cst.Name("v")),
                            ]
                        ),
                        iter=cst.Call(
                            func=cst.Attribute(
                                value=cst.BinaryOperation(
                                    left=cst.Name("_headers"),
                                    operator=cst.BitOr(),
                                    right=cst.Dict(
                                        elements=[
                                            cst.DictElement(
                                                key=cst.SimpleString('"Content-Type"'),
                                                value=cst.SimpleString(
                                                    '"application/json"'
                                                ),
                                            )
                                        ]
                                    ),
                                ),
                                attr=cst.Name("items"),
                            ),
                            args=[],
                        ),
                        ifs=[
                            cst.CompIf(
                                test=cst.Comparison(
                                    left=cst.Name("v"),
                                    comparisons=[
                                        cst.ComparisonTarget(
                                            operator=cst.IsNot(),
                                            comparator=cst.Name("None"),
                                        )
                                    ],
                                )
                            )
                        ],
                    ),
                )
            else:
                headers_expr = cst.Dict(
                    elements=[
                        cst.DictElement(
                            key=cst.SimpleString('"Content-Type"'),
                            value=cst.SimpleString('"application/json"'),
                        )
                    ]
                )
        elif header_params:
            # Filter None values from headers
            headers_expr = cst.DictComp(
                key=cst.Name("k"),
                value=cst.Name("v"),
                for_in=cst.CompFor(
                    target=cst.Tuple(
                        elements=[
                            cst.Element(cst.Name("k")),
                            cst.Element(cst.Name("v")),
                        ]
                    ),
                    iter=cst.Call(
                        func=cst.Attribute(
                            value=cst.Name("_headers"),
                            attr=cst.Name("items"),
                        ),
                        args=[],
                    ),
                    ifs=[
                        cst.CompIf(
                            test=cst.Comparison(
                                left=cst.Name("v"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            )
                        )
                    ],
                ),
            )

        # Params dict
        if query_params:
            params_expr = cst.Name("_params")
        else:
            params_expr = cst.Dict(elements=[])

        # Content (body)
        content_expr: cst.BaseExpression = cst.Name("None")
        if operation.request_body:
            if operation.request_body.content_type == ContentType.JSON:
                if self._dataclass_backend is not None:
                    encode_expr = (
                        self._dataclass_backend.generate_encode_json_expression(
                            operation.request_body.type_annotation or "Any",
                            cst.Name("body"),
                        )
                    )
                    if operation.request_body.required:
                        content_expr = encode_expr
                    else:
                        content_expr = cst.IfExp(
                            test=cst.Comparison(
                                left=cst.Name("body"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            ),
                            body=encode_expr,
                            orelse=cst.Name("None"),
                        )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_request"))],
                    value=cst.Call(
                        func=cst.Name("RequestData"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("method"),
                                value=cst.SimpleString(
                                    f'"{operation.method.value.upper()}"'
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("url"),
                                value=url_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("params"),
                                value=params_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("content"),
                                value=content_expr,
                            ),
                        ],
                    ),
                )
            ]
        )

    def _build_hooked_execution(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build execution with hook function calls baked in."""
        assert self._hooks_config is not None

        statements: list[cst.BaseStatement] = []

        # Pre-hook: _request = await pre_hook(self._inner, _request, _info)
        if self._hooks_config.pre_hook:
            assert isinstance(self._hooks_config.pre_hook, str)
            pre_hook_func = self._hooks_config.get_hook_function(
                self._hooks_config.pre_hook
            )
            code = f"_request = await {pre_hook_func}(self._inner, _request, _info)"
            statements.append(cst.parse_statement(code))

        # Execute HTTP request with async context manager
        exec_code = """
_session = self._ensure_session()
async with _session.request(
    method=_request.method,
    url=_request.url,
    headers=_request.headers,
    params={k: v for k, v in _request.params.items() if v is not None},
    data=_request.content,
) as _resp:
    _content = await _resp.read()
    _response = ResponseData(
        status_code=_resp.status,
        headers=dict(_resp.headers),
        content=_content,
    )
"""
        exec_module = cst.parse_module(exec_code.strip())
        statements.extend(exec_module.body)  # type: ignore[arg-type]

        # Post-hook: _response = await post_hook(self._inner, _response, _info)
        if self._hooks_config.post_hook:
            assert isinstance(self._hooks_config.post_hook, str)
            post_hook_func = self._hooks_config.get_hook_function(
                self._hooks_config.post_hook
            )
            code = f"_response = await {post_hook_func}(self._inner, _response, _info)"
            statements.append(cst.parse_statement(code))

        return statements

    def _build_direct_execution(
        self,
        operation: OperationMethod,
        query_params: list[OperationParameter],
        header_params: list[OperationParameter],
    ) -> list[cst.BaseStatement]:
        """Build direct HTTP execution without hooks."""
        statements: list[cst.BaseStatement] = []

        # Build URL: self._base_url.rstrip("/") + _path
        url_code = '_url = self._base_url.rstrip("/") + _path'
        statements.append(cst.parse_statement(url_code))

        # Build headers
        if header_params:
            if (
                operation.request_body
                and operation.request_body.content_type == ContentType.JSON
            ):
                headers_code = '_headers_final = {k: v for k, v in ({**_headers, "Content-Type": "application/json"}).items() if v is not None}'
            else:
                headers_code = "_headers_final = {k: v for k, v in _headers.items() if v is not None}"
            statements.append(cst.parse_statement(headers_code))
            headers_var = "_headers_final"
        else:
            if (
                operation.request_body
                and operation.request_body.content_type == ContentType.JSON
            ):
                headers_var = '{"Content-Type": "application/json"}'
            else:
                headers_var = "{}"

        # Build query params filter
        if query_params:
            params_expr = "{k: v for k, v in _params.items() if v is not None}"
        else:
            params_expr = "{}"

        # Build content
        content_expr = "None"
        if operation.request_body:
            if operation.request_body.content_type == ContentType.JSON:
                if self._dataclass_backend is not None:
                    # Get the encode expression as code
                    encode_expr = (
                        self._dataclass_backend.generate_encode_json_expression(
                            operation.request_body.type_annotation or "Any",
                            cst.Name("body"),
                        )
                    )
                    encode_code = cst.Module(body=[]).code_for_node(encode_expr)
                    if operation.request_body.required:
                        content_expr = encode_code
                    else:
                        content_expr = f"{encode_code} if body is not None else None"

        # Build async request call with context manager
        request_code = f'''
_session = self._ensure_session()
async with _session.request(
    method="{operation.method.value.upper()}",
    url=_url,
    headers={headers_var},
    params={params_expr},
    data={content_expr},
) as _resp:
    _content = await _resp.read()
    _response = ResponseData(
        status_code=_resp.status,
        headers=dict(_resp.headers),
        content=_content,
    )
'''
        exec_module = cst.parse_module(request_code.strip())
        statements.extend(exec_module.body)  # type: ignore[arg-type]

        return statements

    def _build_path_expression(
        self,
        operation: OperationMethod,
    ) -> cst.BaseExpression:
        """Build the path expression with parameter substitution."""
        path_params = [
            p for p in operation.parameters if p.location == ParameterLocation.PATH
        ]

        if not path_params:
            return cst.SimpleString(f'"{operation.path.template}"')

        # Build f-string with _encode_path calls
        fstring_parts: list[cst.BaseFormattedStringContent] = []
        template = operation.path.template
        last_end = 0

        import re

        for match in re.finditer(r"\{([^}]+)\}", template):
            # Add text before parameter
            if match.start() > last_end:
                fstring_parts.append(
                    cst.FormattedStringText(value=template[last_end : match.start()])
                )

            # Find the parameter
            param_api_name = match.group(1)
            python_name = None
            for p in path_params:
                if p.api_name == param_api_name:
                    python_name = p.name
                    break
            python_name = python_name or param_api_name

            # Add formatted expression with _encode_path
            fstring_parts.append(
                cst.FormattedStringExpression(
                    expression=cst.Call(
                        func=cst.Name("encode_path"),
                        args=[cst.Arg(value=cst.Name(python_name))],
                    )
                )
            )
            last_end = match.end()

        # Add remaining text
        if last_end < len(template):
            fstring_parts.append(cst.FormattedStringText(value=template[last_end:]))

        return cst.FormattedString(parts=fstring_parts)

    def _build_query_params(
        self,
        params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build query parameters dict."""
        dict_elements: list[cst.DictElement] = []
        for param in params:
            dict_elements.append(
                cst.DictElement(
                    key=cst.SimpleString(f'"{param.api_name}"'),
                    value=cst.Name(param.name),
                )
            )

        return cst.SimpleStatementLine(
            body=[
                cst.AnnAssign(
                    target=cst.Name("_params"),
                    annotation=cst.Annotation(
                        annotation=cst.Subscript(
                            value=cst.Name("dict"),
                            slice=[
                                cst.SubscriptElement(
                                    slice=cst.Index(value=cst.Name("str"))
                                ),
                                cst.SubscriptElement(
                                    slice=cst.Index(value=cst.Name("Any"))
                                ),
                            ],
                        )
                    ),
                    value=cst.Dict(elements=dict_elements),
                )
            ]
        )

    def _build_header_params(
        self,
        params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build headers dict."""
        dict_elements: list[cst.DictElement] = []
        for param in params:
            dict_elements.append(
                cst.DictElement(
                    key=cst.SimpleString(f'"{param.api_name}"'),
                    value=cst.Name(param.name),
                )
            )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_headers"))],
                    value=cst.Dict(elements=dict_elements),
                )
            ]
        )

    def _get_success_response(
        self,
        operation: OperationMethod,
    ) -> tuple[str, OperationResponse] | None:
        """Get the first success response (2xx) from operation responses.

        Prioritizes: 200, 201, 202, then any other 2xx, then 'default'.
        """

        # Priority order for success codes
        priority_codes = ["200", "201", "202", "203", "204", "205", "206"]

        for code in priority_codes:
            if code in operation.responses:
                return (code, operation.responses[code])

        # Check for any other 2xx
        for code, response in operation.responses.items():
            if code.startswith("2"):
                return (code, response)

        # Fall back to 'default' if present
        if "default" in operation.responses:
            return ("default", operation.responses["default"])

        return None

    def _build_response_handling(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build response handling code using ResponseData."""
        statements: list[cst.BaseStatement] = []

        # Headers are already dict in ResponseData (both hooked and direct execution paths)
        headers_expr: cst.BaseExpression = cst.Attribute(
            value=cst.Name("_response"),
            attr=cst.Name("headers"),
        )

        # Check for error response: if _response.status_code >= 400
        statements.append(
            cst.If(
                test=cst.Comparison(
                    left=cst.Attribute(
                        value=cst.Name("_response"),
                        attr=cst.Name("status_code"),
                    ),
                    comparisons=[
                        cst.ComparisonTarget(
                            operator=cst.GreaterThanEqual(),
                            comparator=cst.Integer("400"),
                        )
                    ],
                ),
                body=cst.IndentedBlock(
                    body=[
                        cst.SimpleStatementLine(
                            body=[
                                cst.Raise(
                                    exc=cst.Call(
                                        func=cst.Name("APIError"),
                                        args=[
                                            cst.Arg(
                                                keyword=cst.Name("status_code"),
                                                value=cst.Attribute(
                                                    value=cst.Name("_response"),
                                                    attr=cst.Name("status_code"),
                                                ),
                                            ),
                                            cst.Arg(
                                                keyword=cst.Name("body"),
                                                value=cst.Attribute(
                                                    value=cst.Name("_response"),
                                                    attr=cst.Name("content"),
                                                ),
                                            ),
                                            cst.Arg(
                                                keyword=cst.Name("headers"),
                                                value=headers_expr,
                                            ),
                                        ],
                                    )
                                )
                            ]
                        )
                    ]
                ),
                orelse=None,
            )
        )

        # Get success response with priority handling
        success_result = self._get_success_response(operation)

        # Build decode and return statements
        decode_return_stmts: list[cst.BaseStatement] = []

        if success_result:
            status_code, success_response = success_result

            # Handle 204 No Content specially
            if status_code == "204":
                decode_return_stmts.append(self._build_return_none(headers_expr))
            elif success_response.type_annotation:
                # Decode response
                value_expr = self._build_decode_expression(
                    success_response.type_annotation,
                    success_response.content_type,
                )
                decode_return_stmts.append(
                    cst.SimpleStatementLine(
                        body=[
                            cst.Assign(
                                targets=[cst.AssignTarget(target=cst.Name("_result"))],
                                value=value_expr,
                            )
                        ]
                    )
                )

                # Apply on_result hook if configured
                if self._hooks_config is not None and self._hooks_config.on_result:
                    assert isinstance(self._hooks_config.on_result, str)
                    on_result_func = self._hooks_config.get_hook_function(
                        self._hooks_config.on_result
                    )
                    code = (
                        f"_result = await {on_result_func}(self._inner, _result, _info)"
                    )
                    decode_return_stmts.append(cst.parse_statement(code))

                decode_return_stmts.append(self._build_return_result(headers_expr))
            else:
                # No type annotation - return None
                decode_return_stmts.append(self._build_return_none(headers_expr))
        else:
            # No response defined - return None
            decode_return_stmts.append(self._build_return_none(headers_expr))

        # Wrap in try/except if on_error hook is configured
        if self._hooks_config is not None and self._hooks_config.on_error:
            assert isinstance(self._hooks_config.on_error, str)
            on_error_func = self._hooks_config.get_hook_function(
                self._hooks_config.on_error
            )

            # Build except block
            except_code = f"""
except Exception as _e:
    await {on_error_func}(self._inner, _e, _info)
    raise
"""
            except_module = cst.parse_module(except_code.strip())
            except_handler = except_module.body[0]

            # Wrap decode/return in try block
            statements.append(
                cst.Try(
                    body=cst.IndentedBlock(body=decode_return_stmts),
                    handlers=[except_handler.handlers[0]],  # type: ignore[attr-defined]
                    orelse=None,
                    finalbody=None,
                )
            )
        else:
            statements.extend(decode_return_stmts)

        return statements

    def _build_decode_expression(
        self,
        type_annotation: str,
        content_type: ContentType | None,
    ) -> cst.BaseExpression:
        """Build the decode expression based on content type."""
        # bytes expression: _response.content
        bytes_expr = cst.Attribute(
            value=cst.Name("_response"),
            attr=cst.Name("content"),
        )

        if content_type == ContentType.TEXT:
            # Decode bytes to text: _response.content.decode("utf-8")
            return cst.Call(
                func=cst.Attribute(
                    value=bytes_expr,
                    attr=cst.Name("decode"),
                ),
                args=[cst.Arg(value=cst.SimpleString('"utf-8"'))],
            )
        elif content_type == ContentType.BINARY:
            # Return bytes directly: _response.content
            return bytes_expr
        else:
            # JSON (default) - use dataclass backend's decode
            if self._dataclass_backend is not None:
                return self._dataclass_backend.generate_decode_json_expression(
                    type_annotation, bytes_expr
                )
            else:
                # Fallback to json.loads
                return cst.Call(
                    func=cst.Attribute(
                        value=cst.Name("json"),
                        attr=cst.Name("loads"),
                    ),
                    args=[cst.Arg(value=bytes_expr)],
                )

    def _build_return_result(
        self,
        headers_expr: cst.BaseExpression,
    ) -> cst.SimpleStatementLine:
        """Build return statement with ResponseValue wrapper using _result."""
        return cst.SimpleStatementLine(
            body=[
                cst.Return(
                    value=cst.Call(
                        func=cst.Name("ResponseValue"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("value"),
                                value=cst.Name("_result"),
                            ),
                            cst.Arg(
                                keyword=cst.Name("status_code"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("status_code"),
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                        ],
                    )
                )
            ]
        )

    def _build_return_none(
        self,
        headers_expr: cst.BaseExpression,
    ) -> cst.SimpleStatementLine:
        """Build return statement with ResponseValue[None]."""
        return cst.SimpleStatementLine(
            body=[
                cst.Return(
                    value=cst.Call(
                        func=cst.Name("ResponseValue"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("value"),
                                value=cst.Name("None"),
                            ),
                            cst.Arg(
                                keyword=cst.Name("status_code"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("status_code"),
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                        ],
                    )
                )
            ]
        )

    def _get_return_type(self, operation: OperationMethod) -> str:
        """Get the return type annotation for an operation."""
        success_result = self._get_success_response(operation)

        if success_result:
            status_code, success_response = success_result

            # 204 No Content always returns None
            if status_code == "204":
                return "ResponseValue[None]"

            if success_response.type_annotation:
                return f"ResponseValue[{success_response.type_annotation}]"

        return "ResponseValue[None]"

    def get_imports(self) -> list[cst.SimpleStatementLine]:
        """Return required imports for aiohttp."""
        return [
            cst.SimpleStatementLine(
                body=[cst.Import(names=[cst.ImportAlias(name=cst.Name("aiohttp"))])]
            ),
        ]

    def get_base_import_names(self) -> list[str]:
        """ResponseData is always needed since aiohttp uses context manager pattern."""
        return ["ResponseData"]

    def get_stdlib_imports(self) -> list[cst.SimpleStatementLine]:
        """No stdlib imports needed for hooks."""
        return []

    def supports_async(self) -> bool:
        """aiohttp supports async."""
        return True

    def supports_sync(self) -> bool:
        """aiohttp does not support sync."""
        return False

    def get_sub_client_init_args(self) -> list[tuple[str, str]]:
        """Return args for sub-client initialization.

        For aiohttp, we pass the parent client reference so sub-clients can
        access the session after it's created in __aenter__.
        """
        return [("parent", "self"), ("base_url", "self._base_url")]

    def supports_websocket(self) -> bool:
        """aiohttp supports WebSocket natively."""
        return True

    def generate_websocket_method(
        self,
        operation: WebSocketOperation,
    ) -> cst.FunctionDef:
        """Generate a WebSocket connection method.

        Returns a method that creates a WebSocketConnection for the given endpoint.
        """
        method_name = operation.python_method_name()
        params = self._build_websocket_params(operation)
        body = self._build_websocket_method_body(operation)
        return_type = self._get_websocket_return_type(operation)

        decorators: list[cst.Decorator] = []
        if operation.deprecated:
            decorators.append(
                cst.Decorator(
                    decorator=cst.Call(
                        func=cst.Name("deprecated"),
                        args=[
                            cst.Arg(
                                value=cst.SimpleString(
                                    f'"{operation.operation_id} is deprecated"'
                                )
                            )
                        ],
                    )
                )
            )

        # Build docstring from summary/description
        docstring_parts: list[str] = []
        if operation.summary:
            docstring_parts.append(operation.summary)
        if operation.description:
            docstring_parts.append(operation.description)
        docstring = "\n\n".join(docstring_parts) if docstring_parts else None

        func = cst.FunctionDef(
            name=cst.Name(method_name),
            params=params,  # params is now cst.Parameters
            body=cst.IndentedBlock(
                body=([make_docstring(docstring)] + body if docstring else body)
            ),
            returns=make_annotation(return_type),
            decorators=decorators if decorators else [],
        )

        return func

    def _build_websocket_params(
        self,
        operation: WebSocketOperation,
    ) -> cst.Parameters:
        """Build parameter list for a WebSocket connection method."""
        params: list[cst.Param] = [cst.Param(name=cst.Name("self"))]

        # Add path parameters first (required, no default)
        for param in operation.parameters:
            if param.location == ParameterLocation.PATH:
                params.append(
                    make_param(
                        param.name,
                        param.type_annotation,
                    )
                )

        # Add query parameters as keyword-only
        query_params = [
            p for p in operation.parameters if p.location == ParameterLocation.QUERY
        ]
        kwonly_params: list[cst.Param] = []
        for param in query_params:
            default = cst.Name("None") if not param.required else None
            kwonly_params.append(
                make_param(
                    param.name,
                    param.type_annotation,
                    default=default,
                )
            )

        if kwonly_params:
            return cst.Parameters(
                params=params,
                star_arg=cst.ParamStar(),
                kwonly_params=kwonly_params,
            )
        else:
            return cst.Parameters(params=params)

    def _build_websocket_method_body(
        self,
        operation: WebSocketOperation,
    ) -> list[cst.BaseStatement]:
        """Build the method body for a WebSocket connection method."""
        statements: list[cst.BaseStatement] = []

        # Build URL path: _path = f"/ws/chat/{room_id}"
        path_params = {
            p.name: p.api_name
            for p in operation.parameters
            if p.location == ParameterLocation.PATH
        }
        fstring = operation.path.to_python_fstring(
            {v: k for k, v in path_params.items()}
        )
        statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[cst.AssignTarget(target=cst.Name("_path"))],
                        value=cst.FormattedString(
                            parts=self._parse_fstring_parts(fstring),
                        ),
                    )
                ]
            )
        )

        # Build query params dict
        query_params = [
            p for p in operation.parameters if p.location == ParameterLocation.QUERY
        ]

        if query_params:
            # _params = {k: v for k, v in {"key": value, ...}.items() if v is not None}
            dict_elements: list[cst.DictElement] = []
            for param in query_params:
                dict_elements.append(
                    cst.DictElement(
                        key=cst.SimpleString(f'"{param.api_name}"'),
                        value=cst.Name(param.name),
                    )
                )

            statements.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[cst.AssignTarget(target=cst.Name("_params"))],
                            value=cst.DictComp(
                                key=cst.Name("k"),
                                value=cst.Name("v"),
                                for_in=cst.CompFor(
                                    target=cst.Tuple(
                                        elements=[
                                            cst.Element(value=cst.Name("k")),
                                            cst.Element(value=cst.Name("v")),
                                        ]
                                    ),
                                    iter=cst.Call(
                                        func=cst.Attribute(
                                            value=cst.Dict(elements=dict_elements),
                                            attr=cst.Name("items"),
                                        ),
                                        args=[],
                                    ),
                                    ifs=[
                                        cst.CompIf(
                                            test=cst.Comparison(
                                                left=cst.Name("v"),
                                                comparisons=[
                                                    cst.ComparisonTarget(
                                                        operator=cst.IsNot(),
                                                        comparator=cst.Name("None"),
                                                    )
                                                ],
                                            )
                                        )
                                    ],
                                ),
                            ),
                        )
                    ]
                )
            )
        else:
            statements.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[cst.AssignTarget(target=cst.Name("_params"))],
                            value=cst.Name("None"),
                        )
                    ]
                )
            )

        # Return WebSocketConnection(...)
        ws_args: list[cst.Arg] = [
            cst.Arg(
                keyword=cst.Name("session"),
                value=cst.Attribute(
                    value=cst.Name("self"),
                    attr=cst.Name("_session"),
                ),
            ),
            cst.Arg(
                keyword=cst.Name("base_url"),
                value=cst.Attribute(
                    value=cst.Name("self"),
                    attr=cst.Name("_base_url"),
                ),
            ),
            cst.Arg(
                keyword=cst.Name("path"),
                value=cst.Name("_path"),
            ),
            cst.Arg(
                keyword=cst.Name("params"),
                value=cst.Name("_params"),
            ),
        ]

        # Add send_type and receive_type if available
        if operation.send_type:
            ws_args.append(
                cst.Arg(
                    keyword=cst.Name("send_type"),
                    value=cst.Name(operation.send_type),
                )
            )
        if operation.receive_type:
            ws_args.append(
                cst.Arg(
                    keyword=cst.Name("receive_type"),
                    value=cst.Name(operation.receive_type),
                )
            )

        statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Return(
                        value=cst.Call(
                            func=cst.Name("WebSocketConnection"),
                            args=ws_args,
                        )
                    )
                ]
            )
        )

        return statements

    def _get_websocket_return_type(self, operation: WebSocketOperation) -> str:
        """Get the return type annotation for a WebSocket operation."""
        send_type = operation.send_type or "dict[str, Any]"
        receive_type = operation.receive_type or "dict[str, Any]"
        return f"WebSocketConnection[{send_type}, {receive_type}]"

    def _parse_fstring_parts(
        self,
        fstring: str,
    ) -> list[cst.BaseFormattedStringContent]:
        """Parse an f-string into CST parts.

        This is a simplified parser that handles {expr} placeholders.
        """
        import re

        parts: list[cst.BaseFormattedStringContent] = []
        pattern = re.compile(r"\{([^}]+)\}")

        last_end = 0
        for match in pattern.finditer(fstring):
            # Add text before the placeholder
            if match.start() > last_end:
                text = fstring[last_end : match.start()]
                parts.append(cst.FormattedStringText(value=text))

            # Add the placeholder expression
            expr_str = match.group(1)
            # Parse the expression string into CST
            expr = cst.parse_expression(expr_str)
            parts.append(cst.FormattedStringExpression(expression=expr))

            last_end = match.end()

        # Add any remaining text
        if last_end < len(fstring):
            text = fstring[last_end:]
            parts.append(cst.FormattedStringText(value=text))

        return parts

    def get_websocket_runtime_code(self) -> str:
        """Return runtime code for WebSocket support in aiohttp."""
        return '''
# WebSocket support
from typing import TypeVar, Generic, AsyncIterator, Any

TSend = TypeVar("TSend")
TRecv = TypeVar("TRecv")


class WebSocketConnection(Generic[TSend, TRecv]):
    """Typed WebSocket connection wrapper.

    Provides type-safe send/receive operations with automatic
    serialization for the aiohttp backend.
    """

    def __init__(
        self,
        session: aiohttp.ClientSession | None,
        base_url: str,
        path: str,
        params: dict[str, str] | None,
        send_type: type[TSend] | None = None,
        receive_type: type[TRecv] | None = None,
    ) -> None:
        self._session = session
        self._base_url = base_url
        self._path = path
        self._params = params or {}
        self._send_type = send_type
        self._receive_type = receive_type
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self._owned_session: aiohttp.ClientSession | None = None

    async def open(self) -> "WebSocketConnection[TSend, TRecv]":
        """Open the WebSocket connection."""
        if self._session is None:
            # Create our own session if none provided
            self._owned_session = aiohttp.ClientSession(base_url=self._base_url)
            session = self._owned_session
        else:
            session = self._session

        self._ws = await session.ws_connect(
            self._path,
            params=self._params,
        )
        return self

    async def close(self, code: int = 1000, message: bytes = b"") -> None:
        """Close the WebSocket connection."""
        if self._ws is not None:
            await self._ws.close(code=code, message=message)
            self._ws = None
        if self._owned_session is not None:
            await self._owned_session.close()
            self._owned_session = None

    @property
    def closed(self) -> bool:
        """Check if connection is closed."""
        return self._ws is None or self._ws.closed

    async def send(self, message: TSend) -> None:
        """Send a typed message (serialized to JSON)."""
        if self._ws is None:
            raise WebSocketError("WebSocket not connected")
        # Use msgspec if available, otherwise json
        try:
            import msgspec
            data = msgspec.json.encode(message)
            await self._ws.send_bytes(data)
        except ImportError:
            import json
            await self._ws.send_json(message if isinstance(message, dict) else message.__dict__)

    async def receive(self) -> TRecv:
        """Receive and decode a typed message."""
        if self._ws is None:
            raise WebSocketError("WebSocket not connected")

        msg = await self._ws.receive()

        if msg.type == aiohttp.WSMsgType.TEXT:
            data = msg.data
        elif msg.type == aiohttp.WSMsgType.BINARY:
            data = msg.data
        elif msg.type == aiohttp.WSMsgType.CLOSE:
            raise WebSocketClosed(msg.data, str(msg.extra) if msg.extra else "")
        elif msg.type == aiohttp.WSMsgType.ERROR:
            raise WebSocketError(f"WebSocket error: {self._ws.exception()}")
        else:
            raise WebSocketError(f"Unexpected message type: {msg.type}")

        # Decode based on receive_type
        if self._receive_type is not None:
            try:
                import msgspec
                decoder = msgspec.json.Decoder(self._receive_type)
                if isinstance(data, str):
                    data = data.encode("utf-8")
                return decoder.decode(data)
            except ImportError:
                import json
                return json.loads(data)
        else:
            import json
            if isinstance(data, bytes):
                data = data.decode("utf-8")
            return json.loads(data)

    async def send_json(self, data: dict[str, Any]) -> None:
        """Send raw JSON data (untyped)."""
        if self._ws is None:
            raise WebSocketError("WebSocket not connected")
        await self._ws.send_json(data)

    async def receive_json(self) -> dict[str, Any]:
        """Receive raw JSON data (untyped)."""
        if self._ws is None:
            raise WebSocketError("WebSocket not connected")
        msg = await self._ws.receive()
        if msg.type in (aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY):
            return msg.json()
        elif msg.type == aiohttp.WSMsgType.CLOSE:
            raise WebSocketClosed(msg.data, str(msg.extra) if msg.extra else "")
        else:
            raise WebSocketError(f"Unexpected message type: {msg.type}")

    async def __aenter__(self) -> "WebSocketConnection[TSend, TRecv]":
        return await self.open()

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    def __aiter__(self) -> AsyncIterator[TRecv]:
        return self

    async def __anext__(self) -> TRecv:
        try:
            return await self.receive()
        except WebSocketClosed:
            raise StopAsyncIteration


class WebSocketClosed(Exception):
    """Raised when WebSocket connection is closed."""

    def __init__(self, code: int | None = None, reason: str = "") -> None:
        self.code = code
        self.reason = reason
        super().__init__(f"WebSocket closed: {code} {reason}")


class WebSocketError(Exception):
    """General WebSocket error."""
    pass
'''
